/*********************************************************************
 * Desktop Station shield 2 Sketch
 *
 * Copyright (C) 2021 DesktopStation
 *
 */

#define DEBUG

// include the library code:
#include <EEPROM.h>
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM_MM2.h"
#include "DSCoreM.h"
#include "Wire.h"
#include "Functions.h"


#define MAX_S88DECODER 1
#define RELPYERROR_300 "300 Command error"
#define RELPYERROR_301 "301 Syntax error"
#define RELPYERROR_302 "302 receive timeout"
#define RELPYERROR_303 "303 Unknown error"
#define RELPYERROR_NONE ""

#define ENABLE_RAILCOM 1	//0: Disable, 1: Enable

#define LEDSTATE_STOP 0
#define LEDSTATE_RUN 1
#define LEDSTATE_ERR 2
#define LEDSTATE_ANALOG 4

#define PIN_PWMA 21
#define PIN_PWMB 20
#define PIN_RUNLED	2


#define THRESHOLD_EDC 375  // 1024d * (1.2 / 5.9) * 9V / 5V
#define THRESHOLD_OV  999  // 1024d * (1.2 / 5.9) * 20V / 5V

/* Decralation classes */
DSCoreLib DSCore;

/* LED functions */
byte gLED_State = LEDSTATE_STOP;
byte gLED_Toggle = 0;
byte gLED_Counter = 0;


//UART用
String inputString = "";         // a String to hold incoming data


/* Parser */
String gRequest;
String function;
word arguments[4];
word numOfArguments;
byte numOfLocAddr = 0;
word gLocAddresses[4] = {0,0,0,0};
boolean result;


//Task Schedule
unsigned long gPreviousTime_50ms = 0; //     50ms interval
unsigned long gPreviousTime_200ms = 0; //   200ms interval
unsigned long gPreviousTime_1000ms = 0; // 1000ms interval


void SetAnalogSpeed(word inSpeed, byte inDir);
void printString(const char *s, int x, int y);
void ReplyPowerPacket(byte inPower);
void changeLED();


/***********************************************************************
* 
* Boot setup
* 
* 
************************************************************************/

void setup()
{
	//Pin mode
	pinMode(A1, INPUT);//Edc
	pinMode(PIN_PWMA, OUTPUT);//PWM_A
	pinMode(PIN_PWMB, OUTPUT);//PWM_B
	
    pinMode(PIN_RUNLED, OUTPUT);
    digitalWrite(PIN_RUNLED,LOW);
	
	/* Debug message for PC */
	Serial.begin(115200);
	inputString.reserve(200);
	
	Serial.println("-------------------------------------");
	Serial.println("Desktop Station shield 2040");
	Serial.println("-------------------------------------");
	Serial.println("100 Ready");

	DSCore.Init();
	DSCore.gCutOut = ENABLE_RAILCOM;

	
	
	/* LED turn on */
	changeLED();
	
	//Reset task
	gPreviousTime_50ms = millis();
	gPreviousTime_200ms = millis();
	gPreviousTime_1000ms = millis();
	
}


boolean parse() 
{
	int aPosSlash = -1;

	int lpar = gRequest.indexOf('(');
	if (lpar == -1) 
	{
		return false;
	}

	function = String(gRequest.substring(0, lpar));
	function.trim();

	int offset = lpar + 1;
	int aAdddrOffset = 0;
	int comma = gRequest.indexOf(',', offset);
	numOfArguments = 0;
	numOfLocAddr = 0;
	
	while (comma != -1) 
	{
		String tmp = gRequest.substring(offset, comma);
		tmp.trim();
		
		/* スラッシュを探索する */
		aPosSlash = tmp.indexOf('/', 0);
		
		if( aPosSlash > 0)
		{
			tmp = tmp + "/";
			
			//Dammy値を指定
			arguments[numOfArguments++] = 0;
			aAdddrOffset = 0;
			
			//スラッシュはアドレス部だけに許可
			while( aPosSlash != -1)
			{
				String tmp2 = tmp.substring(aAdddrOffset, aPosSlash);
				tmp2.trim();
				
				if( numOfLocAddr < 4)
				{
					gLocAddresses[numOfLocAddr] = stringToWord(tmp2);
					numOfLocAddr++;
				}
				
				aAdddrOffset = aPosSlash + 1;
				aPosSlash = tmp.indexOf('/', aAdddrOffset);
			}
			
			offset = comma + 1;
			comma = gRequest.indexOf(',', offset);
		}
		else
		{
			arguments[numOfArguments++] = stringToWord(tmp);
			offset = comma + 1;
			comma = gRequest.indexOf(',', offset);
		}
	}

	int rpar = gRequest.indexOf(')', offset);
	while (rpar == -1) 
	{
		return false;
	}

	if (rpar > offset) 
	{
		String tmp = gRequest.substring(offset, rpar);
		tmp.trim();
		arguments[numOfArguments++] = stringToWord(tmp);
	}

	return true;
}

boolean dispatch() 
{
	byte aValue;
	byte aValueHigh;
	boolean aResult;
	char aTempStr[16];
	word aLocProtocol;
	byte aDir = 0;
	long aSpeed = 0;
	
	if ((function == "setLocoDirection") || ((function == "sLD")))
	{
		
		if( arguments[1] > 0)
		{
			//FWD=1,REV=2を FWD=0,REV=1に変換
			aDir = arguments[1] - 1;
			
			//異常値はREV固定
			if(aDir > 1)
			{
				aDir = 1;
			}
		}
		else
		{
			//FWD固定
			aDir = 0;
		}
		
		return DSCore.SetLocoDirection(arguments[0], aDir);
	
	}
	else if ((function == "setLocoFunction") || ((function == "sLF")))
	{
		return DSCore.SetLocoFunction(arguments[0], arguments[1] + 1, arguments[2]);
	
	}
	else if ((function == "setTurnout") || ((function == "sTO")))
	{
		return DSCore.SetTurnout(arguments[0], (byte)arguments[1]);
	
	}
	else if ((function == "setPower") || ((function == "sPW")))
	{
		
		if( arguments[0] == 1)
		{
			/* 状態変更 */
			gLED_State = LEDSTATE_RUN;
		}
		else
		{
			/* 状態変更 */
			gLED_State = LEDSTATE_STOP;
		}
		
		
		/* LED状態の変更 */
		changeLED();
		
		return DSCore.SetPower(arguments[0]);
	
	}
	else if ((function == "setLocoSpeed") || ((function == "sLS")))
	{
		return DSCore.SetLocoSpeedEx(arguments[0], arguments[1], arguments[2]);
	}
  else if (function == "DC")
  {
    //PRM0=Speed(0-1023), PRM1=Dir(0,1=FWD, 2=REV)
    if( (gLED_State == LEDSTATE_STOP) || (gLED_State == LEDSTATE_ANALOG))
    {
      
      if( arguments[0] > 0)
      {
        gLED_State = LEDSTATE_ANALOG;
      
        SetAnalogSpeed(arguments[0], arguments[1]);
      }
      else
      {
        SetAnalogSpeed(0, 0);
        gLED_State = LEDSTATE_STOP;
      }
    }
    
    return true;
  }	
	else if ((function == "reset") || ((function == "sRS")))
	{
		Serial.println("100 Ready");
		return true;
	}
	/* reset */
	else if ((function == "setPing") || ((function == "sPI")))
	{
		/* Type(2bytes), Version(2bytes), UID(4bytes) */
		Serial.println("@DSG,00,03,00,02,00,00,00,00");
		return true;
	}
	else if ((function == "getLocoConfig") || ((function == "gCV")))
	{
		/* LED turn on */
		gpio_put(PIN_RUNLED, HIGH);


		aValue = 0;
		aResult = DSCore.ReadConfig( arguments[1], &aValue, 1, 0);
		
		
		//Send to Serial
		Serial.print("@CV,");
		Serial.print(arguments[0]);
		Serial.print(",");
		Serial.print(arguments[1]);
		Serial.print(",");
		
		if( aResult == true)
		{
			Serial.print(aValue);
		}
		else
		{
			Serial.print(0xFFFF, HEX);
		}
		
		Serial.println(",");
		
		/* LED turn off */
		changeLED();
		
		return true;
	}
	else if (function == "getS88")
	{
		int aMaxS88Num = MAX_S88DECODER;
		
		if( arguments[0] > 0)
		{
			aMaxS88Num = arguments[0];
		}
		
		//Send a S88 sensor reply 
		Serial.print("@S88,0000");
		
		/*
		reporter.refresh(aMaxS88Num);

		word aFlags = 0;

		for( int j = 0; j < aMaxS88Num; j++)
		{
		  aFlags = (reporter.getByte((j << 1) + 1) << 8) + reporter.getByte(j << 1);
		  
		  Serial.print((word)aFlags, HEX);
		  Serial.print(",");
		}
		*/
			
		Serial.println("");
		
		return true;
	} /* getS88 */
	else if ((function == "setLocoConfig") || ((function == "sCV")))
	{
		Serial.print("@CVWrite,");
		Serial.print(arguments[1]);
		Serial.print(",");
		Serial.print(arguments[2]);
		Serial.println("");

		/* LED turn on */
		gpio_put(PIN_RUNLED, HIGH);

		DSCore.WriteConfig_Dir( arguments[1], arguments[2]);

		/* LED turn off */
		changeLED();
		
		return true;

	}
	else if ((function == "cutout") || ((function == "CO")))
	{
		/* Railcom CutOut */
		DSCore.gCutOut = (arguments[0] == 1) ? 1: 0;
		
		return true;
	}

	else
	{
		return false;
	}
}

void loop()
{
	uint8_t aErrNo = 0;
	word aEdc;
	
	aErrNo = 0;
	
	//パルスジェネレータのスキャン
	DSCore.Scan();	
	
	//50ms Task
	if( (millis() - gPreviousTime_50ms) >= 50)
	{
		
		//Reset task
		gPreviousTime_50ms = millis();
	}
	
	//200ms Task
	if( (millis() - gPreviousTime_200ms) >= 200)
	{
		
		//LED Control
		taskLED();
		
		/* Sensor check */
		aEdc = analogRead(A1);
		
		/* OV check */
		if( aEdc > THRESHOLD_OV )
		{
			aErrNo = 2;
		}
		/* LV check */
		if( aEdc < THRESHOLD_EDC )
		{
			aErrNo = 1;
			
		}		
		/* エラー処理 */
		if( aErrNo > 0)
		{
			if( DSCore.IsPower() == 1)
			{
				PowerOffByErr(aErrNo);
			}
			
		}
	
		//Reset task
		gPreviousTime_200ms = millis();
	}
	
	
	//1000ms Task
	if( (millis() - gPreviousTime_1000ms) >= 1000)
	{
		
		//Serial.println("@TESTREPLY,9999");
		//Reset task
		gPreviousTime_1000ms = millis();
	}
	
}


void PowerOffByErr(uint8_t inErrNo)
{
	DSCore.SetPower(0);
	delay(1000);
	ReplyPowerPacket(0);
	
	/* 状態変更 */
	gLED_State = LEDSTATE_ERR;
}

void ReplyPowerPacket(byte inPower)
{
	/* Power */
	Serial.print("@PWR,");
	Serial.print(inPower, HEX);
	Serial.println(",");
}


void changeLED()
{
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		gpio_put( PIN_RUNLED, HIGH );
		break;
		
	case LEDSTATE_RUN:
		gpio_put(PIN_RUNLED, HIGH);
		break;
		
	case LEDSTATE_ANALOG:
		break;
	}
	
	return;
}

void taskLED()
{
	/* カウンタを回す(200msで1カウント) */
	gLED_Counter++;
	gLED_Counter = gLED_Counter & 0b1111;
	
	//LED操作
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		
		if( gLED_Counter <= 3)
		{
			gpio_put(PIN_RUNLED, HIGH);
		}
		else
		{
			gpio_put(PIN_RUNLED, LOW);
		}
		break;
		
	case LEDSTATE_RUN:
		break;
		
	case LEDSTATE_ERR:
		
		if( gLED_Toggle == 0)
		{
			gpio_put(PIN_RUNLED, LOW);
		}
		else
		{
			gpio_put(PIN_RUNLED, HIGH);
		}
		
		gLED_Toggle = (~gLED_Toggle) & 1;
		
		break;
	}
}

void SetAnalogSpeed(word inSpeed, byte inDir)
{
  
  word aSpeed = inSpeed >> 2;
  
  if( aSpeed > 255)
  {
    aSpeed = 255;
  }
  
  if( inSpeed == 0)
  {
    FreeAnalogSpeed();
  }
  else
  {
    if( inDir == 2)
    {
      //REV
      analogWrite(PIN_PWMA, 255 - aSpeed);
      analogWrite(PIN_PWMB, 255);
    }
    else
    {
      //FWD
      analogWrite(PIN_PWMB, 255 - aSpeed);
      analogWrite(PIN_PWMA, 255);
    }
  }
}

void FreeAnalogSpeed()
{
  analogWrite(PIN_PWMB, 0);
  analogWrite(PIN_PWMA, 0);
}

//---------------------------------------------------------------------
// シリアル割込み
//---------------------------------------------------------------------

void serialEvent()
{
	//digitalWrite(PIN_DEBUG_MANAGE,HIGH);
	while (Serial.available())
	{
		// get the new byte:
		char inChar = (char)Serial.read();
		
		if (inChar == '\n') {
			
			gRequest = inputString;
			inputString = "";
			
			Serial.println(gRequest);
			
			String aReplyText;
			
	  		if (parse()) {
				if (dispatch()) {
					aReplyText = "200 Ok";
				} else {
					//Serial.println(function);
					aReplyText = RELPYERROR_300;
				}
	  			
			} else {
				aReplyText = RELPYERROR_301;
			}
			
			/* Reply to Desktop Station */
			Serial.println(aReplyText);			
			
			//digitalWrite(PIN_DEBUG_MANAGE,LOW);
			break;
		}
		
		// add it to the inputString:
		inputString += inChar;
	}

}


